import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.HashMap;

import javax.swing.AbstractButton;
import javax.swing.JButton;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JToggleButton;

import javazoom.jlgui.basicplayer.BasicPlayer;
import javazoom.jlgui.basicplayer.BasicPlayerEvent;
import javazoom.jlgui.basicplayer.BasicPlayerException;
import javazoom.jlgui.basicplayer.BasicPlayerListener;

import comp202.fall2007.a5.util.Song;

public class COMP202PlayStatePanel extends JPanel implements PlayStateComponent {
	// *** CONSTANTS

	// Spacing constant
	private static final int SPACING = 5;

	// Play control buttons
	private static final int NUMBER_PLAY_CONTROL_BUTTONS = 5;
	private static final int PREVIOUS = 0;
	private static final int PLAY = 1;
	private static final int PAUSE = 2;
	private static final int STOP = 3;
	private static final int NEXT = 4;

	// Play control button action commands
	private static final String[] PLAY_CONTROL_ACTION_COMMANDS = { "previous",
			"play", "pause", "stop", "next", };

	// Play control button labels
	private static final String[] PLAY_CONTROL_BUTTON_LABELS = { "Previous",
			"Play", "Pause", "Stop", "Next" };

	private static final String OPEN_ERROR = "Could not open track file.";
	private static final String PLAY_ERROR = "Could not play track.";
	private static final String STOP_ERROR = "Could not stop player.";
	private static final String PAUSE_ERROR = "Could not pause play.";
	private static final String RESUME_ERROR = "Could not resume play.";
	private static final String SET_VOLUME_ERROR = "Could not set volume.";
	private static final String SET_BALANCE_ERROR = "Could not set balance.";

	// *** INSTANCE VARIABLES

	// The music-playing object and its state
	private BasicPlayer player;
	private PlayerState playerState;

	// Play control buttons
	private AbstractButton[] playControlButtons;

	// Other components from which we might require some information
	// or that we might need to update due to event on this component
	private PlaylistComponent playlist;
	private PlayAttributesComponent playAttributesControl;
	private TrackInfoComponent trackInfo;

	// *** CONSTRUCTOR

	public COMP202PlayStatePanel(BasicPlayer player, PlaylistComponent playlist) {
		HashMap<Integer, ActionListener> listeners;
		BasicPlayerListener playerListener;

		// Initialize the player and its state
		this.player = player;
		this.playerState = PlayerState.STOP;

		// Initialize the other components.
		if (playlist == null) {
			throw new NullPointerException();
		}
		this.playlist = playlist;
		this.playAttributesControl = null;
		this.trackInfo = null;

		// Set up GUI
		this.playControlButtons = new AbstractButton[NUMBER_PLAY_CONTROL_BUTTONS];
		this.setLayout(new GridLayout(1, NUMBER_PLAY_CONTROL_BUTTONS, SPACING,
				0));
		for (int i = 0; i < this.playControlButtons.length; i++) {
			if (i == PAUSE) {
				this.playControlButtons[i] = new JToggleButton(
						PLAY_CONTROL_BUTTON_LABELS[i]);
				this.playControlButtons[i].setSelected(false);
			} else {
				this.playControlButtons[i] = new JButton(
						PLAY_CONTROL_BUTTON_LABELS[i]);
			}
			this.playControlButtons[i]
					.setActionCommand(PLAY_CONTROL_ACTION_COMMANDS[i]);
			this.playControlButtons[i].setEnabled(false);
			this.add(this.playControlButtons[i]);
		}

		// Set up the event listeners for both the buttons...
		this.playControlButtons[PREVIOUS]
				.addActionListener(new PreviousButtonListener());
		this.playControlButtons[PLAY]
				.addActionListener(new PlayButtonListener());
		this.playControlButtons[STOP]
				.addActionListener(new StopButtonListener());
		this.playControlButtons[PAUSE]
				.addActionListener(new PauseButtonListener());
		this.playControlButtons[NEXT]
				.addActionListener(new NextButtonListener());

		// And the player
		listeners = new HashMap<Integer, ActionListener>();
		listeners.put(BasicPlayerEvent.PLAYING, new PlayEventListener());
		listeners.put(BasicPlayerEvent.STOPPED, new StopEventListener());
		listeners.put(BasicPlayerEvent.PAUSED, new PauseEventListener());
		listeners.put(BasicPlayerEvent.RESUMED, new ResumeEventListener());
		listeners.put(BasicPlayerEvent.EOM, new EOMEventListener());
		playerListener = new TimerPlayerListener(listeners);
		this.player.addBasicPlayerListener(playerListener);
	}

	// *** BUTTON EVENT LISTENERS

	// Play button
	private class PlayButtonListener implements ActionListener {
		public void actionPerformed(ActionEvent event) {
			play();
		}
	}

	// Stop button
	private class StopButtonListener implements ActionListener {
		public void actionPerformed(ActionEvent event) {
			stop();
		}
	}

	// Pause button
	private class PauseButtonListener implements ActionListener {
		public void actionPerformed(ActionEvent event) {
			pause();
		}
	}

	// Previous button
	private class PreviousButtonListener implements ActionListener {
		public void actionPerformed(ActionEvent event) {
			previous();
		}
	}

	// Next button
	private class NextButtonListener implements ActionListener {
		public void actionPerformed(ActionEvent event) {
			next();
		}
	}

	// *** PLAYER EVENT LISTENERS
	// *** These are called from the event-handling thread

	// Play event
	private class PlayEventListener implements ActionListener {
		public void actionPerformed(ActionEvent event) {
			Song track;

			if ((playAttributesControl == null)
					|| ((playAttributesControl != null) && (playAttributesControl
							.getSeekState() == PlayAttributesComponent.SeekState.NORMAL))) {
				track = playlist.get(playlist.getCurrentPosition());

				if (trackInfo != null) {
					trackInfo.set(track.getTitle(), track.getArtist(), track
							.getAlbum());
				}

				playControlButtons[STOP].setEnabled(true);
				playControlButtons[PAUSE].setEnabled(true);
				playControlButtons[PAUSE].setSelected(false);

				updateState();

				playerState = PlayerState.PLAY;
				if (playAttributesControl != null) {
					playAttributesControl.updateState();
				}
			}
		}
	}

	// Stop event
	private class StopEventListener implements ActionListener {
		public void actionPerformed(ActionEvent event) {
			if ((playAttributesControl == null)
					|| ((playAttributesControl != null) && (playAttributesControl
							.getSeekState() == PlayAttributesComponent.SeekState.NORMAL))) {
				if (trackInfo != null) {
					trackInfo.clear();
				}

				playControlButtons[STOP].setEnabled(false);
				playControlButtons[PAUSE].setEnabled(false);
				playControlButtons[PAUSE].setSelected(false);

				playerState = PlayerState.STOP;
				if (playAttributesControl != null) {
					playAttributesControl.updateState();
				}
			}
		}
	}

	// Pause event
	private class PauseEventListener implements ActionListener {
		public void actionPerformed(ActionEvent event) {
			if ((playAttributesControl == null)
					|| ((playAttributesControl != null) && (playAttributesControl
							.getSeekState() == PlayAttributesComponent.SeekState.NORMAL))) {
				playerState = PlayerState.PAUSE;
			}
		}
	}

	// Resume event
	private class ResumeEventListener implements ActionListener {
		public void actionPerformed(ActionEvent event) {
			if ((playAttributesControl == null)
					|| ((playAttributesControl != null) && (playAttributesControl
							.getSeekState() == PlayAttributesComponent.SeekState.NORMAL))) {
				playerState = PlayerState.PLAY;
			}
		}
	}

	// End-of-media event
	private class EOMEventListener implements ActionListener {
		public void actionPerformed(ActionEvent event) {
			if (!playlist.hasNextTrack()) {
				playerState = PlayerState.STOP;
				playlist.setCurrentPosition(0);
				updateState();
			} else {
				next();
			}
		}
	}

	// *** PRIVATE METHODS

	private double getGain() {
		return ((double) playAttributesControl.getCurrentVolume())
				/ (playAttributesControl.getMaximumVolume() - playAttributesControl
						.getMinimumVolume());
	}

	private double getPan() {
		return ((double) playAttributesControl.getCurrentBalance())
				/ (playAttributesControl.getMaximumAbsoluteBalance() - playAttributesControl
						.getMinimumAbsoluteBalance());
	}

	// *** PUBLIC METHODS SPECIFIED BY THE INTERFACE

	public PlayerState getPlayerState() {
		return this.playerState;
	}

	public void play() {
		Song track;

		try {
			if ((playerState == PlayerState.PLAY)
					|| (playerState == PlayerState.PAUSE)) {
				player.stop();
			}
		} catch (BasicPlayerException exception) {
			JOptionPane.showMessageDialog(this, STOP_ERROR);
			return;
		}

		try {
			track = playlist.get(playlist.getCurrentPosition());
			player.open(track.getFilePath());
		} catch (BasicPlayerException exception) {
			JOptionPane.showMessageDialog(this, OPEN_ERROR);
			if (playlist.hasNextTrack()) {
				next();
				play();
			}
			return;
		}

		try {
			player.play();
		} catch (BasicPlayerException exception) {
			JOptionPane.showMessageDialog(this, PLAY_ERROR);
			if (playlist.hasNextTrack()) {
				next();
				play();
			}
			return;
		}

		try {
			player.setGain(getGain());
		} catch (BasicPlayerException exception) {
			JOptionPane.showMessageDialog(this, SET_VOLUME_ERROR);
		}

		try {
			player.setPan(getPan());
		} catch (BasicPlayerException exception) {
			JOptionPane.showMessageDialog(this, SET_BALANCE_ERROR);
		}
	}

	public void stop() {
		try {
			player.stop();
		} catch (BasicPlayerException exception) {
			JOptionPane.showMessageDialog(this, STOP_ERROR);
		}
	}

	public void pause() {
		if (playerState == PlayerState.PLAY) {
			try {
				player.pause();
			} catch (BasicPlayerException exception) {
				JOptionPane.showMessageDialog(this, PAUSE_ERROR);
				return;
			}
		} else {
			try {
				player.resume();
			} catch (BasicPlayerException exception) {
				JOptionPane.showMessageDialog(this, RESUME_ERROR);
				if (playlist.hasNextTrack()) {
					next();
				}
				return;
			}

			try {
				player.setGain(getGain());
			} catch (BasicPlayerException exception) {
				JOptionPane.showMessageDialog(this, SET_VOLUME_ERROR);
			}

			try {
				player.setPan(getPan());
			} catch (BasicPlayerException exception) {
				JOptionPane.showMessageDialog(this, SET_BALANCE_ERROR);

			}
		}
	}

	public void previous() {
		if (playlist.hasPreviousTrack()) {
			playlist.moveToPreviousTrack();

			if (playerState != PlayerState.STOP) {
				play();
			} else {
				this.playControlButtons[PREVIOUS].setEnabled(playlist
						.hasPreviousTrack());
			}
		} else {
			// Should not happen
			this.playControlButtons[PREVIOUS].setEnabled(false);
		}
	}

	public void next() {
		if (playlist != null && playlist.hasNextTrack()) {
			playlist.moveToNextTrack();

			if (playerState != PlayerState.STOP) {
				play();
			} else {
				this.playControlButtons[NEXT].setEnabled(playlist
						.hasNextTrack());
			}
		} else {
			// Should not happen
			this.playControlButtons[NEXT].setEnabled(false);
		}
	}

	public void updateState() {
		this.playControlButtons[PLAY].setEnabled(!playlist.isEmpty());
		this.playControlButtons[PREVIOUS].setEnabled(playlist
				.hasPreviousTrack());
		this.playControlButtons[NEXT].setEnabled(playlist.hasNextTrack());
	}

	// *** OTHER PUBLIC METHODS

	public PlaylistComponent getPlaylist() {
		return playlist;
	}

	public void setPlaylist(PlaylistComponent playlist) {
		this.playlist = playlist;
	}

	public PlayAttributesComponent getPlayAttributesControl() {
		return playAttributesControl;
	}

	public void setPlayAttributesControl(
			PlayAttributesComponent playAttributesControl) {
		this.playAttributesControl = playAttributesControl;
	}

	public TrackInfoComponent getTrackInfo() {
		return trackInfo;
	}

	public void setTrackInfo(TrackInfoComponent trackInfo) {
		this.trackInfo = trackInfo;
	}
}