import java.awt.BorderLayout;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.NoSuchElementException;

import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableColumn;

import comp202.fall2007.a5.comparator.SongComparator;
import comp202.fall2007.a5.util.Song;

public class COMP202PlaylistPanel extends JPanel implements PlaylistComponent {
	private static final int NUMBER_COLUMNS = 6;
	private static final int MARKER_COLUMN = 0;
	private static final int TRACK_COLUMN = 1;
	private static final int TITLE_COLUMN = 2;
	private static final int ARTIST_COLUMN = 3;
	private static final int ALBUM_COLUMN = 4;
	private static final int LENGTH_COLUMN = 5;

	private final String[] columnNames = { "  ", "#", "Title", "Artist",
			"Album", "Length" };
	private static final String MARKER = ">>";
	private static final String EMPTY = "  ";

	private IteratedPlaylist playlist;
	private DecimalFormat timeFormatter;

	private JTable playlistTable;
	private AbstractTableModel playlistModel;

	private PlayStateComponent playStateControl;
	private PlaylistControlComponent playlistControl;

	public COMP202PlaylistPanel() {
		JScrollPane scrollPane;
		TableColumn column;
		ListSelectionModel selectionModel;
		int columnCount;
		int width;

		this.playlist = new ThreadSafeIteratedPlaylist();
		this.timeFormatter = new DecimalFormat("00");

		this.playStateControl = null;
		this.playlistControl = null;

		this.playlistModel = new PlaylistTableModel();
		this.playlistTable = new JTable(playlistModel);
		this.playlistTable
				.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);

		columnCount = playlistModel.getColumnCount();
		for (int i = 0; i < columnCount; i++) {
			column = this.playlistTable.getColumnModel().getColumn(i);

			width = 0;
			switch (i) {
			case MARKER_COLUMN:
			case TRACK_COLUMN:
				width = 25;
				break;
			case TITLE_COLUMN:
				width = 150;
				break;
			case ARTIST_COLUMN:
			case ALBUM_COLUMN:
				width = 100;
				break;
			case LENGTH_COLUMN:
				width = 50;

			}
			column.setPreferredWidth(width);
		}

		scrollPane = new JScrollPane(playlistTable);
		this.setLayout(new BorderLayout());
		this.add(scrollPane, BorderLayout.CENTER);

		selectionModel = this.playlistTable.getSelectionModel();
		selectionModel
				.addListSelectionListener(new PlaylistSelectionListener());
		this.playlistTable.addMouseListener(new PlaylistClickListener());
	}

	// *** COMPONENT SUPPORT OBJECTS

	// Playlist table model
	private class PlaylistTableModel extends AbstractTableModel {

		public String getColumnName(int col) {
			return columnNames[col];
		}

		public int getColumnCount() {
			return NUMBER_COLUMNS;
		}

		public boolean isCellEditable(int rowIndex, int columnIndex) {
			return false;
		}

		public int getRowCount() {
			return playlist.getSize();
		}

		public Object getValueAt(int rowIndex, int columnIndex) {
			Song song;
			Object value;
			int hours;
			int minutes;
			int seconds;
			int totalSeconds;

			song = playlist.get(rowIndex);
			value = null;
			switch (columnIndex) {
			case MARKER_COLUMN:
				if (rowIndex == playlist.getCurrentPosition()) {
					value = MARKER;
				} else {
					value = EMPTY;
				}
				break;
			case TRACK_COLUMN:
				value = new Integer(song.getTrackNumber());
				break;
			case TITLE_COLUMN:
				value = song.getTitle();
				break;
			case ARTIST_COLUMN:
				value = song.getArtist();
				break;
			case ALBUM_COLUMN:
				value = song.getAlbum();
				break;
			case LENGTH_COLUMN:
				totalSeconds = song.getLength();

				hours = totalSeconds / 3600;
				minutes = (totalSeconds % 3600) / 60;
				seconds = totalSeconds % 60;

				value = ((hours == 0) ? "" : timeFormatter.format(hours) + ":")
						+ timeFormatter.format(minutes) + ":"
						+ timeFormatter.format(seconds);
				break;
			}

			return value;
		}
	}

	// *** LISTENERS

	// Selection of playlist items
	private class PlaylistSelectionListener implements ListSelectionListener {
		public void valueChanged(ListSelectionEvent event) {
			int[] selected;

			selected = playlistTable.getSelectedRows();
			if (playlistControl != null) {
				playlistControl.selectionUpdated(selected);
			}
		}
	}

	// *** MOUSE LISTENERS

	// Mouse events on playlist item
	private class PlaylistClickListener extends MouseAdapter {
		public void mouseClicked(MouseEvent event) {
			int oldCurrent;
			int newCurrent;
			
			if (event.getClickCount() >= 2) {
				oldCurrent = playlist.getCurrentPosition();
				newCurrent = playlistTable.getSelectedRow();
				
				playlist.setCurrentPosition(newCurrent);
				playlistModel.fireTableRowsUpdated(oldCurrent, oldCurrent);
				playlistModel.fireTableRowsUpdated(newCurrent, newCurrent);
				
				if (playStateControl != null) {
					playStateControl.play();
				}
			}
		}
	}

	// *** PUBLIC METHOD REQUIRED BY INTERFACE

	public void add(int index, Song song) {
		this.playlist.add(index, song);

		if (this.playStateControl != null) {
			this.playStateControl.updateState();
		}
		if (this.playlistControl != null) {
			this.playStateControl.updateState();
		}

		this.playlistModel.fireTableRowsInserted(index, index);
	}

	public boolean add(Song song) {
		boolean result;
		int size;

		result = this.playlist.add(song);
		if (result) {
			if (this.playStateControl != null) {
				this.playStateControl.updateState();
			}
			if (this.playlistControl != null) {
				this.playStateControl.updateState();
			}

			size = this.playlist.getSize();
			this.playlistModel.fireTableRowsInserted(size - 1, size - 1);
		}
		return result;
	}

	public boolean addAll(ArrayList<Song> songList) {
		boolean result;
		int oldSize;
		int newSize;

		oldSize = this.playlist.getSize();
		result = this.playlist.addAll(songList);
		if (result) {
			if (this.playStateControl != null) {
				this.playStateControl.updateState();
			}
			if (this.playlistControl != null) {
				this.playStateControl.updateState();
			}

			newSize = this.playlist.getSize();
			this.playlistModel.fireTableRowsInserted(oldSize, newSize - 1);
		}
		return result;
	}

	public boolean addAll(Playlist playList) {
		boolean result;
		int oldSize;
		int newSize;

		oldSize = this.playlist.getSize();
		result = this.playlist.addAll(playList);
		if (result) {
			if (this.playStateControl != null) {
				this.playStateControl.updateState();
			}
			if (this.playlistControl != null) {
				this.playStateControl.updateState();
			}

			newSize = this.playlist.getSize();
			this.playlistModel.fireTableRowsInserted(oldSize, newSize - 1);
		}
		return result;
	}

	public boolean addAll(Song[] songArray) {
		boolean result;
		int oldSize;
		int newSize;

		oldSize = this.playlist.getSize();
		result = this.playlist.addAll(songArray);
		if (result) {
			if (this.playStateControl != null) {
				this.playStateControl.updateState();
			}
			if (this.playlistControl != null) {
				this.playStateControl.updateState();
			}

			newSize = this.playlist.getSize();
			this.playlistModel.fireTableRowsInserted(oldSize, newSize - 1);
		}
		return result;
	}

	public void clear() {
		if ((this.playStateControl != null)
				&& (this.playStateControl.getPlayerState() != PlayStateComponent.PlayerState.STOP)) {
			this.playStateControl.stop();
		}

		this.playlist.clear();
		if (this.playStateControl != null) {
			this.playStateControl.updateState();
		}
		if (this.playlistControl != null) {
			this.playStateControl.updateState();
		}
		this.playlistModel.fireTableDataChanged();
	}

	public Song get(int index) {
		return this.playlist.get(index);
	}

	public int getCurrentPosition() {
		return this.playlist.getCurrentPosition();
	}

	public int getPlaylistSize() {
		return this.playlist.getSize();
	}

	public boolean hasNextTrack() {
		return this.playlist.hasNextTrack();
	}

	public boolean hasPreviousTrack() {
		return this.playlist.hasPreviousTrack();
	}

	public boolean isEmpty() {
		return this.playlist.isEmpty();
	}

	public ArrayList<String> load(File path, MusicCollection collection)
			throws IOException {
		ArrayList<String> badPaths;

		badPaths = this.playlist.load(path, collection);
		if (this.playStateControl != null) {
			this.playStateControl.updateState();
		}
		if (this.playlistControl != null) {
			this.playStateControl.updateState();
		}
		this.playlistModel.fireTableDataChanged();

		return badPaths;
	}

	public void moveDown(int index, int numberPositions) {
		int destination;

		this.playlist.moveDown(index, numberPositions);

		if (this.playStateControl != null) {
			this.playStateControl.updateState();
		}
		if (this.playlistControl != null) {
			this.playStateControl.updateState();
		}

		destination = index + numberPositions;
		playlistModel.fireTableRowsUpdated(index, destination);
		playlistTable.setRowSelectionInterval(destination, destination);
	}

	public void moveToNextTrack() throws NoSuchElementException {
		int current;

		current = this.playlist.getCurrentPosition();
		this.playlist.moveToNextTrack();

		if (this.playStateControl != null) {
			this.playStateControl.updateState();
		}
		playlistModel.fireTableRowsUpdated(current, current + 1);
	}

	public void moveToPreviousTrack() throws NoSuchElementException {
		int current;

		current = this.playlist.getCurrentPosition();
		this.playlist.moveToPreviousTrack();

		if (this.playStateControl != null) {
			this.playStateControl.updateState();
		}
		playlistModel.fireTableRowsUpdated(current - 1, current);
	}

	public void moveUp(int index, int numberPositions) {
		int destination;

		this.playlist.moveUp(index, numberPositions);

		if (this.playStateControl != null) {
			this.playStateControl.updateState();
		}
		if (this.playlistControl != null) {
			this.playStateControl.updateState();
		}

		destination = index - numberPositions;
		playlistModel.fireTableRowsUpdated(destination, index);
		playlistTable.setRowSelectionInterval(destination, destination);
	}

	public Song removeSong(int index) {
		Song song;

		int current;

		current = this.playlist.getCurrentPosition();
		if ((index == current)
				&& (this.playStateControl != null)
				&& (this.playStateControl.getPlayerState() != PlayStateComponent.PlayerState.STOP)) {
			this.playStateControl.stop();
		}

		song = this.playlist.remove(index);

		if (this.playStateControl != null) {
			this.playStateControl.updateState();
		}
		if (this.playlistControl != null) {
			this.playStateControl.updateState();
		}

		this.playlistModel.fireTableRowsDeleted(index, index);

		return song;
	}

	public void save(File path) throws IOException {
		this.playlist.save(path);
	}

	public void setCurrentPosition(int current)
			throws IndexOutOfBoundsException {
		int oldCurrent;

		oldCurrent = this.playlist.getCurrentPosition();
		this.playlist.setCurrentPosition(current);

		if (this.playStateControl != null) {
			this.playStateControl.updateState();
		}
		if (this.playlistControl != null) {
			this.playStateControl.updateState();
		}

		this.playlistModel.fireTableRowsUpdated(oldCurrent, oldCurrent);
		this.playlistModel.fireTableRowsUpdated(current, current);
	}

	public void sort(SongComparator comparator) {
		this.playlist.sort(comparator);

		if (this.playStateControl != null) {
			this.playStateControl.updateState();
		}
		if (this.playlistControl != null) {
			this.playStateControl.updateState();
		}

		this.playlistModel.fireTableDataChanged();
	}

	public PlayStateComponent getPlayStateControl() {
		return playStateControl;
	}

	public void setPlayStateControl(PlayStateComponent playStateControl) {
		this.playStateControl = playStateControl;
	}

	public PlaylistControlComponent getPlaylistControl() {
		return playlistControl;
	}

	public void setPlaylistControl(PlaylistControlComponent playlistControl) {
		this.playlistControl = playlistControl;
	}

	// *** OTHER PUBLIC METHODS
}