package comp202.fall2007.a5.filter;

/**
 * <p>
 * A <code>NumberMatchType</code> object defines a way to determine whether
 * two integer values match. It is used mostly by <code>SongFilter</code>s
 * who define conditions based on numeric attributes of <code>Song</code>s,
 * such as the track number or length, to determine whether a <code>Song</code>
 * matches the conditions they define.
 * </p>
 */
public abstract class NumberMatchType {
	private final String symbol;

	/**
	 * Implements number matching as a comparison for inequality where the value
	 * of the first argument is greater than or equal to the value of the second
	 * argument. It is used in the construction of <code>SongFilter</code>s to
	 * indicate that this matching type should be used by the new
	 * <code>SongFilter</code>. In other words, if a <code>SongFilter</code>
	 * that defines a condition based on a numeric attribute of
	 * <code>Song</code>s uses this <code>NumberMatchType</code>, it
	 * accepts the <code>Song</code> if the value of the <code>Song</code>'s
	 * attribute is greater than or equal to the filter value.
	 */
	public static final NumberMatchType AT_LEAST = new NumberMatchType(">=") {
		public boolean match(int left, int right) {
			return left >= right;
		}
	};

	/**
	 * Implements number matching as a comparison for inequality where the value
	 * of the first argument is less than or equal to the value of the second
	 * argument. It is used in the construction of <code>SongFilter</code>s to
	 * indicate that this matching type should be used by the new
	 * <code>SongFilter</code>. In other words, if a <code>SongFilter</code>
	 * that defines a condition based on a numeric attribute of
	 * <code>Song</code>s uses this <code>NumberMatchType</code>, it
	 * accepts the <code>Song</code> if the value of the <code>Song</code>'s
	 * attribute is less than or equal to the filter value.
	 */
	public static final NumberMatchType AT_MOST = new NumberMatchType("<=") {
		public boolean match(int left, int right) {
			return left <= right;
		}
	};

	/**
	 * Implements number matching as a comparison for equality. It is used in the
	 * construction of <code>SongFilter</code>s to indicate that this
	 * matching type should be used by the new <code>SongFilter</code>. In other
	 * words, if a <code>SongFilter</code> that defines a condition based on a
	 * numeric attribute of <code>Song</code>s uses this
	 * <code>NumberMatchType</code>, it accepts the <code>Song</code> if
	 * the value of the <code>Song</code>'s attribute is equal to the filter
	 * value.
	 */
	public static final NumberMatchType EQUAL = new NumberMatchType("==") {
		public boolean match(int left, int right) {
			return left == right;
		}
	};

	private NumberMatchType(String symbol) {
		this.symbol = symbol;
	}

	/**
	 * <p>
	 * Returns a <code>String</code> representation of this
	 * <code>NumberMatchType</code> which describes the type of matching it
	 * represents.
	 * </p>
	 * 
	 * @return A <code>String</code> representation of this
	 *         <code>NumberMatchType</code>.
	 */
	public String toString() {
		return this.symbol;
	}

	/**
	 * <p>
	 * Determines whether the first specified integer value matches the second
	 * specified integer value according to the matching algorithm implemented
	 * by this <code>NumberMatchType</code>.
	 * </p>
	 * 
	 * <p>
	 * This method is called internally by methods defined in the
	 * <code>SongFilter</code> class, and does not need to be called by
	 * classes which use <code>SongFilter</code>s.
	 * </p>
	 * 
	 * @param left
	 *            The first integer value
	 * @param right
	 *            The second integer value
	 * @return <code>true</code> if <code>left</code> matches
	 *         <code>right</code> according to the type of matching algorithm
	 *         implemented by this <code>NumberMatchType</code>,
	 *         <code>false</code> otherwise.
	 */
	public abstract boolean match(int left, int right);
}
