package comp202.fall2007.a5.filter;

/**
 * <p>
 * A <code>StringMatchType</code> object defines a way to determine whether
 * two <code>String</code>s match. It is used mostly by
 * <code>SongFilter</code>s who define conditions based on text attributes of
 * <code>Song</code>s, such as the artist name or album title, to determine
 * whether a <code>Song</code> matches the conditions they define.
 * </p>
 */
public abstract class StringMatchType {
	private final String symbol;

	/**
	 * <p>
	 * Implements <code>String</code> matching as comparison for equality.
	 * Used in the construction of <code>SongFilter</code>s to indicate that
	 * equality matching should be used by a <code>SongFilter</code>. In
	 * other words, if a <code>SongFilter</code> that defines a condition
	 * based on a text attribute of <code>Song</code>s uses this
	 * <code>StringMatchType</code>, it accepts the <code>Song</code> if
	 * the value of the <code>Song</code>'s attribute is equal to the filter
	 * value.
	 * </p>
	 */
	public static final StringMatchType EQUALS = new StringMatchType("is") {
		public boolean match(String left, String right, boolean ignoreCase) {
			return ignoreCase ? left.equalsIgnoreCase(right) : left
					.equals(right);
		}
	};

	/**
	 * <p>
	 * Implements <code>String</code> matching as substring. Used in the
	 * construction of <code>SongFilter</code>s to indicate that substring
	 * matching should be used by a <code>SongFilter</code>. In other words,
	 * if a <code>SongFilter</code> that defines a condition based on a text
	 * attribute of <code>Song</code>s uses this <code>StringMatchType</code>,
	 * it accepts the <code>Song</code> if the value of the <code>Song</code>'s
	 * attribute contains the filter value as a substring.
	 * </p>
	 */
	public static final StringMatchType CONTAINS = new StringMatchType(
			"contains") {
		public boolean match(String left, String right, boolean ignoreCase) {
			boolean result;
			String upperCaseLeft;
			String upperCaseRight;

			if (ignoreCase) {
				upperCaseLeft = left.toUpperCase();
				upperCaseRight = right.toUpperCase();
				result = (upperCaseLeft.indexOf(upperCaseRight) != -1);
			} else {
				result = (left.indexOf(right) != -1);
			}
			return result;
		}
	};

	private StringMatchType(String symbol) {
		this.symbol = symbol;
	}

	/**
	 * <p>
	 * Returns a <code>String</code> representation of this
	 * <code>StringMatchType</code> which describes the type of matching it
	 * represents.
	 * </p>
	 * 
	 * @return A <code>String</code> representation of this
	 *         <code>StringMatchType</code>.
	 */
	public String toString() {
		return symbol;
	}

	/**
	 * <p>
	 * Determines whether the first specified <code>String</code> matches the
	 * second specified <code>String</code> according to the matching
	 * algorithm implemented by this <code>StringMatchType</code>.
	 * </p>
	 * 
	 * <p>
	 * This method is called internally by methods defined in the
	 * <code>SongFilter</code> class, and does not need to be called by
	 * classes which use <code>SongFilter</code>s.
	 * </p>
	 * 
	 * @param left
	 *            The first <code>String</code>
	 * @param right
	 *            The second <code>String</code>
	 * @param ignorecase
	 *            <code>true</code> if the alphabetic case of letters should
	 *            be ignored when deciding whether <code>left</code> and
	 *            <code>right</code> match, <code>false</code> otherwise.
	 * @return <code>true</code> if <code>left</code> matches
	 *         <code>right</code> according to the type of matching algorithm
	 *         implemented by this <code>StringMatchType</code>,
	 *         <code>false</code> otherwise.
	 */
	public abstract boolean match(String left, String right, boolean ignorecase);
}
