package comp202.fall2007.a5.util;

import java.io.File;
import java.io.FileFilter;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.blinkenlights.jid3.ID3Exception;
import org.blinkenlights.jid3.MP3File;
import org.blinkenlights.jid3.v1.ID3V1Tag;
import org.blinkenlights.jid3.v1.ID3V1_1Tag;
import org.blinkenlights.jid3.v2.ID3V2Tag;
import org.jaudiotagger.audio.AudioFileIO;
import org.jaudiotagger.audio.AudioHeader;

import comp202.fall2007.a5.util.file.ExtensionFileFilter;

/**
 * A <code>SongFactory</code> enables the easy creation of <code>Song</code>
 * objects representing digital audio tracks stored on disk. The created
 * <code>Song</code> objects's attributes are initialized using the values
 * stored in the metadata (ID3) tags stored in the file.
 */
public class SongFactory {
	private static final FileFilter MP3_FILTER = new ExtensionFileFilter("mp3");

	// Static initializer: disable the Loggers used by the Jaudiotagger tag
	// libraries
	{
		disableLogging();
	}

	/**
	 * Creates a new <code>SongFactory</code>.
	 */
	public SongFactory() {
	}

	/**
	 * <p>
	 * Creates a new <code>Song</code> object representing the MP3 audio track
	 * stored in the file given by the specified <code>String</code>.
	 * </p>
	 * 
	 * <p>
	 * The length attribute of the new <code>Song</code> object is initialized
	 * from the audio header of the MP3 file, and the disc number attribute is
	 * always <code>1</code>.
	 * </p>
	 * 
	 * <p>
	 * To initialize the artist, album, title and track number attributes of the
	 * new <code>Song</code> object, the method determines whether the MP3
	 * file contains an ID3 V2 tag. If it does, then the aforementioned
	 * attributes of the new <code>Song</code> object are initialized using
	 * the data stored in this tag. Otherwise, the method determines whether the
	 * MP3 file contains an ID3 V1 tag; if it does, the aforementioned
	 * attributes are initialized using the data stored in this tag, otherwise,
	 * they are assigned default values.
	 * </p>
	 * 
	 * <p>
	 * The default value for the artist, album, and title attributes is the
	 * empty <code>String</code>, while the default value for the track
	 * number attribute is <code>Song.EMPTY_TRACK</code>. Note that if a tag
	 * is present, but one of the fields corresponding to the artist, album,
	 * title, or track number attribute is empty, the corresponding attribute in
	 * returned <code>Song</code> object is assigned the default value.
	 * </p>
	 * 
	 * @param path
	 *            The path of the MP3 file which contains the audio track to be
	 *            represented by the returned <code>Song</code>.
	 * @return A new <code>Song</code> objects whose attributes are
	 *         initialized using values read from the file given by
	 *         <code>path</code>.
	 * @throws IOException
	 *             if there was an error reading the file specified by
	 *             <code>path</code>.
	 * @throws FileNotFoundException
	 *             if the file specified by <code>path</code> does not exist.
	 * @throws IncorrectFileExtensionException
	 *             if the file specified by <code>path</code> does not have
	 *             extension .MP3 (or case variations thereof).
	 */
	public Song createSong(String path) throws IOException,
			FileNotFoundException, IncorrectFileExtensionException {
		IOException wrapperException;
		AudioHeader audioHeader;
		MP3File songFile;
		ID3V2Tag v2Tag;
		ID3V1Tag v1Tag;
		String title;
		String artist;
		String album;
		File file;
		int trackNumber;
		int length;

		artist = "";
		album = "";
		title = "";
		trackNumber = Song.EMPTY_TRACK;
		length = 0;

		file = new File(path);
		if (!file.exists()) {
			throw new FileNotFoundException(path);
		}

		if (MP3_FILTER.accept(file)) {
			throw new IncorrectFileExtensionException(path);
		}

		try {
			audioHeader = AudioFileIO.read(file).getAudioHeader();
			length = audioHeader.getTrackLength();
		} catch (Exception exception) {
			wrapperException = new IOException();
			wrapperException.initCause(exception);
			throw wrapperException;
		}

		songFile = new MP3File(file);
		v2Tag = null;
		try {
			v2Tag = songFile.getID3V2Tag();
		} catch (ID3Exception exception) {
		}

		if (v2Tag != null) {
			artist = v2Tag.getArtist();
			if (artist == null) {
				artist = "";
			}
			album = v2Tag.getAlbum();
			if (album == null) {
				album = "";
			}
			title = v2Tag.getTitle();
			if (title == null) {
				title = "";
			}
			try {
				trackNumber = v2Tag.getTrackNumber();
			} catch (ID3Exception exception) {
			}
		} else {
			v1Tag = null;
			try {
				v1Tag = songFile.getID3V1Tag();
			} catch (ID3Exception exception) {
			}

			if (v1Tag != null) {
				artist = v1Tag.getArtist();
				if (artist == null) {
					artist = null;
				}
				album = v1Tag.getAlbum();
				if (album == null) {
					album = null;
				}
				title = v1Tag.getTitle();
				if (title == null) {
					title = null;
				}
				if (v1Tag instanceof ID3V1_1Tag) {
					trackNumber = ((ID3V1_1Tag) v1Tag).getAlbumTrack();
				}
			}
		}

		return new Song(file, artist, album, title, trackNumber, length);
	}

	/**
	 * Enable the <code>Logger</code>s used and created by the Jaudiotagger
	 * libraries.
	 * 
	 * Logging is disabled by default.
	 */
	public static void enableLogging() {
		Logger logger;
		Handler[] handlers;

		logger = Logger.getLogger("");
		handlers = logger.getHandlers();
		for (Handler handler : handlers) {
			handler.setLevel(Level.OFF);
		}
	}

	/**
	 * Disable the <code>Loggers</code> used and created by the Jaudiotagger
	 * libraries.
	 * 
	 * Logging is disabled by default.
	 */
	public static void disableLogging() {
		Logger logger;
		Handler[] handlers;

		logger = Logger.getLogger("");
		handlers = logger.getHandlers();
		for (Handler handler : handlers) {
			handler.setLevel(Level.OFF);
		}
	}
}
